from flask import Flask, jsonify, request, Response
from helpers.loader import Loader
from helpers.threads import Threads
from helpers.env import read_config

app = Flask(__name__)
env = read_config()


class Server:

    @staticmethod
    def setup():
        pass

    @staticmethod
    def run():
        Server.setup()
        app.run(port=int(env.get('global', 'server_port', fallback=5000)))

    @staticmethod
    @app.route('/')
    def g_index():
        return jsonify({"title": "Mybi-ci REST API",
                        "urls": [
                            "/ (GET)",
                            "/run (POST)",
                            "/log/<build_id>/<log_file> (GET)"
                        ]})

    @staticmethod
    @app.route('/run', methods=['POST'])
    def p_run():
        build = request.get_json()
        task = Loader.build_root_task(build)
        if task:
            Threads.run_detached(task.run)
            return jsonify({
                "log_file": task.log_file,
                "build_id": task.build_id,
                "task": task.id
            })
        else:
            return jsonify({
                "Error": "Load error",
                "Cause": "Your build json is not valid"
            })

    @staticmethod
    @app.route('/log/<build_id>/<log_file>', methods=['GET'])
    def g_log(build_id, log_file):
        with open(env.get('global', 'log_dir')+"/"+build_id+"/"+log_file) as f:
            content = f.readlines()
        return Response(content, mimetype='text/plain')



__author__ = 'marcoantonioalberoalbero'

import logging
import uuid
import os


class Log:

    config = None
    consolidate_only = None

    def __init__(self):
        self.default_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    def setup(self, config, task_id, build_id=None, log_dir=None, log_file=None, with_console=True):
        """
        Setup the default logger for the task_id in question
        :param config: config environment
        :param task_id: id of the task whose logger is going to be configured
        :param log_dir: the dir where the files are kept (None is config default)
        :param log_file: the log file (None is config default)
        :return:
        """
        Log.config = config
        Log.consolidate_only = True if config.get('global', 'log_consolidate_only') == "True" else False
        level = (logging.DEBUG if 'debug' in config.get('global', 'log_level') else logging.INFO)
        log_format = (config.get('global', 'log_format')
                      if config.get('global', 'log_format')
                      else self.default_format)
        logger = logging.getLogger(config.get('global', 'program') + '.' + task_id)
        logger.setLevel(level)
        ld = (log_dir if log_dir else config.get('global', 'log_dir'))
        lf = (log_file if log_file else config.get('global', 'log_file'))
        if build_id:
            if not os.path.exists(ld+'/'+build_id):
                os.makedirs(ld+'/'+build_id)
            fh = logging.FileHandler(ld+'/'+build_id+'/'+lf)
        else:
            fh = logging.FileHandler(ld+'/'+lf)
        formatter = logging.Formatter(log_format)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        if with_console:
            ch = logging.StreamHandler()
            ch.setFormatter(formatter)
            logger.addHandler(ch)
        return logger

    @staticmethod
    def id_log(logger=None, msg=None, level=logging.INFO):
        """
        Log with a uuid for track the log of the task/command
        :param logger: logger who's going to log
        :param msg: message to log
        :param level: level to log
        :return:
        """
        if logger and msg and level == logging.INFO:
            logger.info(str(uuid.uuid1()) + " - " + msg)
        if logger and msg and level == logging.DEBUG:
            logger.debug(str(uuid.uuid1()) + " - " + msg)

    @staticmethod
    def set_log(task, l_dir=None, l_file=None, consolidate_only=True, with_console=True):
        """
        Adds the logger with logger_id new loggers to log with
        :param task: the task to configure the log for
        :param l_dir: log directory
        :param l_file: log file
        :param consolidate_only: if only consolidate the log or not
        :return:
        """
        if consolidate_only:
            # log to the consolidated file (here it breaks with its own logger, it's not part of the logger hierarchy)
            # so it only logs in this l_file.
            tid = 'consolidated.' + task.id
        else:
            # add a new child logger (because it's part of the logger hierarchy)
            tid = task.id
        task.log = Log().setup(config=Log.config,
                               task_id=tid,
                               build_id=task.build_id,
                               log_dir=l_dir,
                               log_file=l_file,
                               with_console=with_console)

    @staticmethod
    def consolidate_log(task, hierarchy_node=None):
        """
        Consolidate the log of the task to every task ancestor
        :param task: task whose log is going to be consolidated
        :param hierarchy_node: ancestor
        :return:
        """
        if not Log.consolidate_only:
            if hierarchy_node:
                Log.set_log(task, consolidate_only=False, l_file=hierarchy_node.log_file, with_console=False)
                if hierarchy_node.parent:
                    Log.consolidate_log(task, hierarchy_node=hierarchy_node.parent)



__author__ = 'marcoantonioalberoalbero'

from jsonschema import validate, Draft4Validator


class JsonValidator:
    def __init__(self):
        pass

    @staticmethod
    def validate(json, schema):
        result = {}
        v = Draft4Validator(schema)
        err = sorted(v.iter_errors(json), key=str)
        result['errors'] = err
        if not len(err):
            result['valid'] = True
        else:
            result['valid'] = False
        return result



__author__ = 'marcoantonioalberoalbero'

import multiprocessing


class Threads:
    def __init__(self):
        pass

    @staticmethod
    def run_detached(routine):
        p = multiprocessing.current_process()
        name = "worker-" + str(p.pid)
        cc = multiprocessing.Process(name=name, target=routine)
        cc.daemon = False
        cc.start()


__author__ = 'marcoantonioalberoalbero'


class TasksSchema:

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "definitions": {
            "identifier": {
                "type": "string",
                "pattern": "^[A-Za-z0-9\\.-_]+$"
            },
            "task": {
                "type": "object",
                "properties": {
                    "id": {"$ref": "#/definitions/identifier"},
                    "env": {
                        "type": "array",
                        "items": {
                            "property": {
                                "type": "string"
                            },
                            "value": {
                                "type": "string"
                            }
                        }
                    },
                    "parallel_tasks": {
                        "type": "array",
                        "items": {
                            "anyOf": [
                                {
                                    "$ref": "#/definitions/task"
                                },
                                {
                                    "$ref": "#/definitions/command"
                                }
                            ]
                        }
                    },
                    "sequential_tasks": {
                        "type": "array",
                        "items": {
                            "anyOf": [
                                {
                                    "$ref": "#/definitions/task"
                                },
                                {
                                    "$ref": "#/definitions/command"
                                }
                            ]
                        }
                    }
                },
                "required": [
                    "id"
                ],
                "anyOf": [
                    {
                        "required": ["parallel_tasks"]
                    },
                    {
                        "required": ["sequential_tasks"]
                    }
                ],
                "additionalProperties": False
            },
            "command": {
                "type": "object",
                "properties": {
                    "id": {"$ref": "#/definitions/identifier"},
                    "env": {
                        "type": "array",
                        "items": {
                            "property": {
                                "type": "string"
                            },
                            "value": {
                                "type": "string"
                            }
                        }
                    },
                    "cmd": {
                        "type": "string"
                    }
                },
                "required": [
                    "id", "cmd"
                ],
                "additionalProperties": False
            }
        },

        "type": "object",
        "properties": {
            "build": {
                "type": "string"
            },
            "starter": {
                "anyOf": [
                    {
                        "$ref": "#/definitions/task"
                    },
                    {
                        "$ref": "#/definitions/command"
                    }
                ]
            }
        },
        "required": ["build", "starter"],
        "additionalProperties": False
    }


__author__ = 'marcoantonioalberoalbero'

import optparse
import os.path
import sys
from helpers.loader import Loader
from helpers.logging import Log
from helpers.env import read_config
from server.server import Server

# Support vars.
ENDL = "\n"
usage = """
\twhere options are:
\t\t -s => server mode, start a server for receiving jobs
\t\t -b build_file.json => build
\t\t -t build_file.json => test config
"""


class Executor:
    """ Class for execute the ci engine """
    def __init__(self):
        self.parser = optparse.OptionParser(usage='usage: %prog [options]'+usage)
        self.parser.add_option('-b', '--build', action='store_true', default=False, help='build mode')
        self.parser.add_option('-t', '--test', action='store_true', default=False, help='test mode')
        self.parser.add_option('-s', '--server', action='store_true', default=False, help='server mode')
        self.config = read_config()
        self.logger = Log().setup(config=self.config, task_id=__name__)

    def run(self, args):
        (options, arguments) = self.parser.parse_args(args)

        if options.server:
            Server.run()
            sys.exit(0)

        if (options.test or options.build) and len(args) == 2:
            file_path = args[1]
            op = "Testing " if options.test else "Building "
            if os.path.isfile(file_path):
                Log.id_log(self.logger, op + args[1])
                task = Loader.load_from_file(file_path)
            else:
                Log.id_log(self.logger, str(args[1]) + " is not a valid file")
                sys.exit(0)
            if options.build:
                Log.id_log(self.logger, "Running " + args[1])
                task.run()
                Log.id_log(self.logger, "Your build ID: "
                           + os.environ['MYBICI_BUILD_ID'] +
                           " has finished with status: " + str(task.finish_status))
            sys.exit(0)

        self.parser.print_usage()



__author__ = 'marcoantonioalberoalbero'

import os
import subprocess
import datetime
from helpers.logging import Log


class Command:
    def __init__(self, config, name):
        self.parent = None
        self.name = name
        self.config = config
        self.consolidate_only = True if config.get('global', 'log_consolidate_only') == "True" else False
        self.command = None
        self.finish_status = None
        self.env = dict(os.environ)
        self.id = __name__ + '.' + name
        self.start_time = datetime.datetime.now()
        self.log_file = self.id + '-' + self.start_time.strftime("%Y-%m-%d_%H:%M:%S") + '.log'
        # log to its own file
        self.build_id = os.getenv('MYBICI_BUILD_ID', None)
        self.log = Log().setup(config=config, task_id=self.id, build_id=self.build_id, log_file=self.log_file)

    def set_command(self, command):
        self.command = command

    def add_env_var(self, key, value):
        self.env[key] = os.path.expandvars(value)

    def run(self, result_queue=None):
        if not self.consolidate_only:
            Log.consolidate_log(task=self, hierarchy_node=self.parent)
        else:
            Log.set_log(self, consolidate_only=True)

        Log.id_log(self.log, "running command " + self.command)
        result = subprocess.Popen(self.command,
                                  shell=True,
                                  env=self.env,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  universal_newlines=True)

        # Poll process for new output until finished
        while True:
            line = result.stdout.readline()
            self.finish_status = result.poll()
            if isinstance(self.finish_status, int) and not line:
                break
            Log.id_log(self.log, line)
        if result_queue:
            result_queue.put(self.finish_status)
        Log.id_log(self.log, "Task " + self.name + " finished with status: " + str(self.finish_status))
        return self.finish_status



__author__ = 'marcoantonioalberoalbero'

from queue import Queue
import threading
import datetime
from helpers.logging import Log
from helpers.constants import Constants
import os


class Task:
    def __init__(self, config, name):
        self.parent = None
        self.name = name
        self.finish_status = Constants.CMD_OK
        self.config = config
        self.env = dict(os.environ)
        self.consolidate_only = True if config.get('global', 'log_consolidate_only') == "True" else False
        self.parallel_tasks = Queue()
        self.sequential_tasks = Queue()
        self.id = __name__ + '.' + name
        self.start_time = datetime.datetime.now()
        self.log_file = self.id+'-'+self.start_time.strftime("%Y-%m-%d_%H:%M:%S")+'.log'
        # log to its own file
        self.build_id = os.getenv('MYBICI_BUILD_ID', None)
        self.log = Log().setup(config=config, task_id=self.id, build_id=self.build_id, log_file=self.log_file)

    def is_runnable(self, task):
        op = getattr(task, "run", None)
        if op and callable(op):
            return True
        return False

    def add_task(self, task, sequential=True):
        if self.is_runnable(task):
            task.parent = self
            if sequential:
                self.sequential_tasks.put(task)
            else:
                self.parallel_tasks.put(task)

    def add_env_var(self, key, value):
        self.env[key] = os.path.expandvars(value)
        # set it up in the os environment in order to be passed to the child tasks
        os.environ[key] = os.path.expandvars(value)

    def run(self, result_queue=None, stop_on_error=True):
        parallels = []
        if not self.consolidate_only:
            Log.consolidate_log(task=self, hierarchy_node=self.parent)
        else:
            Log.set_log(self, consolidate_only=True)

        Log.id_log(self.log, "running task " + self.name)
        t_queue = Queue()
        while not self.parallel_tasks.empty():
            task = self.parallel_tasks.get()
            t = threading.Thread(target=task.run, args=[t_queue])
            parallels.append(t)
            t.start()
        while not self.sequential_tasks.empty():
            task = self.sequential_tasks.get()
            task.run()
            if task.finish_status != Constants.CMD_OK:
                self.finish_status = task.finish_status
                if stop_on_error:
                    break
        for t in parallels:
            t.join()
            result = t_queue.get_nowait()
            if result != Constants.CMD_OK:
                self.finish_status = result

        if result_queue:
            result_queue.put(self.finish_status)
        Log.id_log(self.log, "Task " + self.name + " finished with status: " + str(self.finish_status))
        return self.finish_status



__author__ = 'marcoantonioalberoalbero'

import configparser
import os


def read_config():
    global config, sections
    try:
        init = os.environ['MYBICI_SETUP_FILE']
        sections = os.environ['MYBICI_SETUP_FILE_SECTIONS'].split(',')
        config = configparser.ConfigParser()
        config.read(init)
    except:
        raise Exception("Need to define MYBICI_SETUP_FILE and "
                        "MYBICI_SETUP_FILE_SECTIONS environment "
                        "variables (https://github.com/periket2000/mybi-ci)")
    return config



__author__ = 'marcoantonioalberoalbero'
from helpers.task import Task
from commands.command import Command
from helpers.env import read_config
from validators.json_validator import JsonValidator
from validators.tasks_schema import TasksSchema
from helpers.logging import Log
import json
import os
import uuid


class Loader:

    env = read_config()
    logger = Log().setup(config=env, task_id=__name__)

    @staticmethod
    def load_from_file(file):
        with open(file) as data_file:
            build = json.load(data_file)
            task = Loader.build_root_task(build)
            if not task:
                Log.id_log(Loader.logger, "Build file format not valid (" + file + ")")
            return task

    @staticmethod
    def build_root_task(build):
        result = JsonValidator.validate(build, TasksSchema.schema)
        if result['valid'] is True:
            os.environ['MYBICI_BUILD_ID'] = str(uuid.uuid1())
            Log.id_log(Loader.logger, "Loading build '" + build["build"] + "'")
            task = Loader.load(build["starter"])
            Log.id_log(Loader.logger, "Load of '" + build["build"] + "' [OK], BUILD ID: " + os.environ['MYBICI_BUILD_ID'])
            return task
        return None

    @staticmethod
    def load(json_data):
        if "cmd" in json_data:
            result = Loader.load_command(json_data)
        else:
            result = Loader.load_task(json_data)
        return result

    @staticmethod
    def load_task(json_data):
        task = Task(Loader.env, json_data["id"])
        if "env" in json_data:
            for entry in json_data["env"]:
                for k in entry.keys():
                    task.add_env_var(k, entry[k])
        if "parallel_tasks" in json_data:
            for t in json_data["parallel_tasks"]:
                if Loader.is_command(t):
                    task.add_task(Loader.load_command(t), sequential=False)
                else:
                    task.add_task(Loader.load_task(t), sequential=False)
        if "sequential_tasks" in json_data:
            for t in json_data["sequential_tasks"]:
                if Loader.is_command(t):
                    task.add_task(Loader.load_command(t))
                else:
                    task.add_task(Loader.load_task(t))
        return task

    @staticmethod
    def load_command(json_data):
        cmd = Command(Loader.env, json_data["id"])
        if "env" in json_data:
            for entry in json_data["env"]:
                for k in entry.keys():
                    cmd.add_env_var(k, entry[k])
        if "cmd" in json_data:
            cmd.set_command(json_data["cmd"])
        return cmd

    @staticmethod
    def is_command(json_data):
        if "cmd" in json_data:
            return True
        else:
            return False
